/* Exercise: timing blocking sends
*
*  In this exercise you will try blocking
*  communications (syncronous and buffered)
*  to send a matrix from process 0 to process 1.
*
* Follow this rules:
*  - Before the two communication, a barrier should
*    syncronize the processes;
*  - Process 0 will begin the blocking send
*  - while process 1 will sleep for 5 seconds 
*    before starting the receive
*  - measure how long process 0 will take in
*    the two different sending calls
*/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>

#define MSIZE 100

int main(int argc, char *argv[]) {

    MPI_Status status;
    int rank, nprocs;
    double *mpibuffer;
    int     mpibuffer_length;

    int i;
    double startTime,endTime;

    /* data to communicate */
    double  vector[MSIZE];
    
    /* Start up MPI */
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    
    /* print banner */
    if (rank == 0) {
        printf("\nCommunication mode examples: process 0 sends a vector\n"); 
        printf("to process 1 using a syncronous and a buffered send.\n");
        printf("Process 1 waits for 5 seconds before starting the\n");
        printf("corresponding receive call. Time spent in the syncronous\n");
        printf("and buffered send is reported.\n\n");

        fflush(stdout);
    }

    /* simple datatype communication examples using two processes */
    if (nprocs < 2) {

        if (rank == 0) {
            printf("\nSORRY: need at least 2 processes.\n");
        }

        /* exit MPI on error */
        MPI_Finalize();
        
        exit(EXIT_FAILURE);
    }

    if (rank == 0) {
        for (i = 0; i < MSIZE; i++)
            vector[i] = (double) i;
    }

    MPI_Barrier(MPI_COMM_WORLD);

    /* SYNCHRONOUS COMMUNICATION */
    if (rank == 0) {
        startTime = MPI_Wtime();

        /* Process 0 start the Synchronous Send */
        MPI_Ssend(vector, MSIZE, MPI_DOUBLE, 1, 666, MPI_COMM_WORLD); 
    
        endTime = MPI_Wtime();

        printf("Proc 0 spent %10.7f to exit from SSend.\n", endTime - startTime);

    }

    if (rank == 1) {
        
        /* Process 1 sleeps for 5 seconds and start the receive */
        sleep(3);

        MPI_Recv(vector, MSIZE, MPI_DOUBLE, 0, 666, MPI_COMM_WORLD, 
            &status);
    }

    if (rank == 0) {
        for (i = 0; i < MSIZE; i++)
            vector[i] = (double) i;
    }

    MPI_Barrier(MPI_COMM_WORLD);

    /* BUFFERED COMMUNICATION */
    if (rank == 0) {

        mpibuffer_length = (MSIZE*sizeof(double) + MPI_BSEND_OVERHEAD);
        mpibuffer = (double *) malloc (mpibuffer_length);
        MPI_Buffer_attach(mpibuffer, mpibuffer_length);

        startTime = MPI_Wtime();

        MPI_Bsend(vector, MSIZE, MPI_DOUBLE, 1, 666, MPI_COMM_WORLD); 
    
        endTime = MPI_Wtime();

        printf("Proc 0 spent %10.7f to exit from Bsend.\n", endTime - startTime);
        MPI_Buffer_detach(mpibuffer, &mpibuffer_length);

    }
    
    if (rank == 1) {
        
        /* Process 1 sleeps for 5 seconds and start the receive */
        sleep(5);

        MPI_Recv(vector, MSIZE, MPI_DOUBLE, 0, 666, MPI_COMM_WORLD, 
            &status);
        for (i=0;i<MSIZE;i++)
	  printf("%10.7f\n",vector[i]);
    }


    /* Quit */
    MPI_Finalize();

    return 0;
}
