program laplace
    
    use mpi

    implicit none

    double precision, dimension(:), allocatable :: T, Tnew, T0
    double precision :: tol, variation=1.0
    integer :: msize, maxIter, i, iter=0
    ! timing
    double precision :: startTime, endTime
    ! MPI variables
    integer :: ierr, rank, nprocs, status(MPI_STATUS_SIZE)
    integer :: tag=201, mymsize
    double precision :: myvariation=1.0
    integer :: cartesianComm, procsgrid(1)=0
    integer :: source_rl, dest_rl, source_lr, dest_lr
    logical :: periods(1)=.FALSE.

    ! MPI init...
    call MPI_Init(ierr)
    call MPI_Comm_rank(MPI_COMM_WORLD, rank, ierr)
    call MPI_Comm_size(MPI_COMM_WORLD, nprocs, ierr) 

    if ( rank == 0 ) then
        write(*,*) 'Enter mesh size, max iterations and tollerance:'
        read(*,*)  msize, maxIter, tol
        ! process 0 allocates the global array T0 for displaying results 
        allocate (T0(msize))
    endif

    startTime = MPI_Wtime()

    ! broadcast input parameters (msize,maxIter,tol) from process 0 to others
    call MPI_Bcast(msize, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(maxIter, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(tol, 1, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)

    if ( mod(msize,nprocs) /= 0 ) then
        if ( rank == 0 ) then
            write(*,*) 'The number of processes must fit the mesh size.'
            write(*,*) 'Mesh size           =', msize
            write(*,*) 'Number of processes =', nprocs
        endif
        call MPI_Finalize(ierr)
        stop
    endif

    ! computing 1D processes grid
    call MPI_Dims_Create(nprocs, 1, procsgrid, ierr)
    if (rank == 0) write(*,*) 'Processes mesh is =', procsgrid

    ! creating a new communicator with a 1D cartesian topology 
    call MPI_Cart_Create(MPI_COMM_WORLD, 1, procsgrid, periods, .FALSE., &
            cartesianComm, ierr)

    ! calculating problem size/process
    mymsize = msize / nprocs

    ! allocating memory for T and Tnew
    allocate (T(0:mymsize+1), Tnew(mymsize))

    ! set up boundary values
    T(0:mymsize+1) = 0.0
    if ( rank == (nprocs-1) ) T(mymsize+1) = 100.0

    ! calculating source/dest neighbours (right->left)
    call MPI_Cart_shift(cartesianComm, 0, -1, source_rl, dest_rl, ierr)
    ! calculating source/dest neighbours (left->right)
    call MPI_Cart_shift(cartesianComm, 0,  1, source_lr, dest_lr, ierr)

    do while (variation > tol .and. iter <= maxIter)

        iter = iter + 1

        !--- exchange boundary data with neighbours (right->left)
        call MPI_Sendrecv(T(1), 1, MPI_DOUBLE_PRECISION, dest_rl, tag,         & 
                        T(mymsize+1), 1, MPI_DOUBLE_PRECISION, source_rl, tag, & 
                        cartesianComm, status, ierr);

        !--- exchange boundary data with neighbours (left->right)
        call MPI_Sendrecv(T(mymsize), 1, MPI_DOUBLE_PRECISION, dest_lr, tag+1,  &
                        T(0), 1, MPI_DOUBLE_PRECISION, source_lr, tag+1,        &
                        cartesianComm, status, ierr);

        ! calculating new solution
        Tnew = 0.5 * ( T(0:mymsize-1) + T(2:mymsize+1) )
        ! calculating the error
        myvariation = maxval(dabs( Tnew - T(1:mymsize) ))

        ! find the maximum values of variation among all processes
        call MPI_Allreduce(myvariation, variation, 1, MPI_DOUBLE_PRECISION, &
                    MPI_MAX, MPI_COMM_WORLD, ierr)

        if( mod(iter,100) == 0 .and. rank == 0 ) &
            write(*,"(a,i8,e12.4)") ' iter, variation:', iter, variation

        ! updating solution
        T(1:mymsize) = Tnew
    enddo

    ! gather data to process 0 (into array T0) for displaying
    call MPI_Gather(T(1), mymsize, MPI_DOUBLE_PRECISION,  &
                    T0, mymsize, MPI_DOUBLE_PRECISION, 0, &
                    MPI_COMM_WORLD, ierr)

    endTime = MPI_Wtime()

    ! output of the results
    if ( rank == 0 ) then
        write(*,'(A,F10.4)') ' Elapsed time (s)     =', endTime - startTime

        write(*,*) 'Mesh size            =', msize
        write(*,*) 'Stopped at iteration =', iter
        write(*,*) 'The maximum error    =', variation
        
        open(10, file="results")
        do i = 1, msize
            write(10, "(i8, e18.9)") i, T0(i)
        end do
        close(10)
    endif
    
    deallocate (T, Tnew)

    call MPI_Finalize(ierr)

end program laplace
